namespace CSLint
{
    using System;
    using Reflector.Disassembler;

    internal class GraphEdge
    {
        internal GraphEdge   next;
        internal Instruction insn;
        internal MethodDesc  methodDesc;
        internal GraphVertex vertex;
        internal CallContext callCtx;
        internal long        mask;
		
        internal GraphEdge(GraphVertex vertex, MethodDesc methodDesc, Instruction insn, GraphEdge sibling, CallContext cc)
        {
            next = sibling;
            callCtx = cc;
            mask = 0;
            this.insn = insn;
            this.vertex = vertex;
            this.methodDesc = methodDesc;
        }
		
        internal void dump(GraphVertex from) 
        {
            Console.Error.WriteLine("Edge from vertex " + from + " to vertex " + vertex + " in method " + methodDesc);
        }

        internal void message(int loopId)
        {
            mask |= (1L << (loopId & 63));
            String msg = "Loop " + loopId + ": locking monitor " + vertex + " in method " + 
                    methodDesc + " can cause deadlock";
            String srcFile = methodDesc.cls.getSourceFilePath();
            if (srcFile != null) 
            {
                CSLint.instance.message(srcFile + ":" + methodDesc.getLineNumber(insn) + ": " + msg);
            } 
            else 
            {
                CSLint.instance.message(msg);
            }
            for (CallContext cc = callCtx; cc.inv != null; cc = cc.up)
            {
                msg = "  Loop " + loopId + ": called from " + cc.inv.caller;
                srcFile = cc.inv.caller.cls.getSourceFilePath();
                if (srcFile != null) 
                { 
                    CSLint.instance.message(srcFile + ":" + cc.inv.caller.getLineNumber(cc.inv.insn) + ": " + msg);
                } 
                else 
                {
                    CSLint.instance.message(msg);
                }
            }
        }
    }
}