namespace CSLint
{
    using System;
    using Reflector.Disassembler;

    public class GraphVertex
    {
        internal GraphEdge   edges;
        internal GraphVertex next;
        internal GraphVertex pred;
        internal GraphEdge   backEdge;
        internal GraphVertex breadthFirst;
        internal int         marker;
        
        internal Object      syncObject;
        internal ClassDesc   syncClass;
        internal ClassDesc   syncMetaclass;

        internal static GraphVertex graph;		
		
        public override String ToString() 
        {
            FieldDesc fd = syncObject as FieldDesc;
            if (fd != null) 
            { 
                return fd.field.DeclaringType + "." + fd.field.Name;
            } 
            MethodDesc md = syncObject as MethodDesc;
            if (md != null)
            { 
                return md.method.DeclaringType + "." + md.method.Name;
            } 
            return syncClass != null 
                ? syncClass.cls.ToString()
                : syncMetaclass != null
                  ? ("typeof(" + syncMetaclass.cls + ")")
                  : "";
        }

        internal static void  verify()
        {
            int nLoops = 0;
            int marker = 1;
            for (GraphVertex root = graph; root != null; marker++, root = root.next)
            {
                root.breadthFirst = null;
                root.marker = marker;
                bool loopFound = false;
                GraphVertex thisLevel = root;
                do 
                {
                    GraphVertex nextLevel = null;
                    for (GraphVertex v = thisLevel; v != null; v = v.breadthFirst)
                    {
                        for (GraphEdge e = v.edges; e != null; e = e.next)
                        {
                            GraphVertex ev = e.vertex;
                            if (ev == root)
                            {
                                long mask = e.mask;
                                GraphVertex p;
                                for (p = v; p != root; p = p.pred)
                                {
                                    mask &= p.backEdge.mask;
                                }
                                loopFound = true;
                                if (mask == 0)
                                {
                                    nLoops += 1;
                                    e.message(nLoops);
                                    for (p = v; p != root; p = p.pred)
                                    {
                                        p.backEdge.message(nLoops);
                                    }
                                }
                            }
                            else if (ev.marker != marker)
                            {
                                ev.pred = v;
                                ev.marker = marker;
                                ev.breadthFirst = nextLevel;
                                nextLevel = ev;
                                ev.backEdge = e;
                            }
                        }
                    }
                    // Make it possible to GC to collection unused contexts
                    while (thisLevel != null)
                    {
                        GraphVertex nextSibling = thisLevel.breadthFirst;
                        thisLevel.breadthFirst = null;
                        thisLevel = nextSibling;
                    }
                    thisLevel = nextLevel;
                }
                while (thisLevel != null && (!loopFound || CSLint.instance.reportAllLoops));
            }   
            if (nLoops == 0) 
            { 
                CSLint.instance.message("No deadlocks found");
            }
        }
		
        internal void addEdge(GraphVertex vertex, MethodDesc caller, Instruction insn, CallContext cc)
        {
            for (GraphEdge edge = edges; edge != null; edge = edge.next)
            {
                if (edge.vertex == vertex && edge.methodDesc == caller && edge.insn.Offset == insn.Offset && (!CSLint.instance.reportAllLoops || edge.callCtx.inv == cc.inv))
                {
                    return;
                }
            }
            edges = new GraphEdge(vertex, caller, insn, edges, cc);
            //edges.dump(this);
        }
		
        public GraphVertex()
        {
            marker = 0;
            next = graph;
            graph = this;
            edges = null;
        }
    }
}